/*--------------------------------------------------------------------------
 *  Copyright 2008 utgenome.org
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *--------------------------------------------------------------------------*/
//--------------------------------------
// utgb-shell Project
//
// UTGBShell.java
// Since: Jan 8, 2008
//
// $URL$ 
// $Author$
//--------------------------------------
package org.utgenome.shell;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.reflect.Modifier;
import java.util.List;
import java.util.TreeMap;

import org.xerial.core.XerialException;
import org.xerial.util.FileResource;
import org.xerial.util.ResourceFilter;
import org.xerial.util.bean.BeanUtil;
import org.xerial.util.cui.OptionParser;
import org.xerial.util.cui.OptionParserException;
import org.xerial.util.io.VirtualFile;
import org.xerial.util.log.Logger;

/**
 * A command line client entry point
 * 
 * @author leo
 * 
 */
public class UTGBShell {

	private static Logger _logger = Logger.getLogger(UTGBShell.class);

	public enum Opt {
		HELP, VERSION
	}

	public static TreeMap<String, UTGBShellSubCommand> subCommandTable = new TreeMap<String, UTGBShellSubCommand>();

	/**
	 * search sub commands from the this package (org.utgenome.shell)
	 */
	public static void searchSubCommands() {
		String shellPackage = UTGBShell.class.getPackage().getName();
		List<VirtualFile> classFileList = FileResource.listResources(shellPackage, new ResourceFilter() {
			public boolean accept(String resourcePath) {
				return resourcePath.endsWith(".class");
			}
		});

		for (VirtualFile vf : classFileList) {
			String logicalPath = vf.getLogicalPath();
			int dot = logicalPath.lastIndexOf(".");
			if (dot <= 0)
				continue;
			String className = shellPackage + "." + logicalPath.substring(0, dot).replaceAll("/", ".");

			try {
				Class c = Class.forName(className, false, UTGBShell.class.getClassLoader());
				if (!Modifier.isAbstract(c.getModifiers()) && UTGBShellSubCommandBase.class.isAssignableFrom(c)) {
					// found a sub command class
					UTGBShellSubCommand subCommand = (UTGBShellSubCommand) c.newInstance();
					if (subCommand == null)
						continue;
					subCommandTable.put(subCommand.name(), subCommand);
				}
			}
			catch (ClassNotFoundException e) {
				continue;
			}
			catch (InstantiationException e) {
				_logger.error(e);
			}
			catch (IllegalAccessException e) {
				_logger.error(e);
			}

		}
	}

	static {
		// search the all available sub commands
		searchSubCommands();
	}

	public static void main(String[] args) {
		OptionParser<Opt> optionParser = new OptionParser<Opt>();
		optionParser.setIgnoreUnknownOption(true);
		optionParser.addOption(Opt.HELP, "h", "help", "display help message");
		optionParser.addOption(Opt.VERSION, "v", "version", "display version");

		try {
			optionParser.parse(args);

			if (optionParser.getArgumentLength() > 0) {
				// go to sub command processing
				String subCommandName = optionParser.getArgument(0);
				UTGBShellSubCommand subCommand = subCommandTable.get(subCommandName);
				if (subCommand != null) {
					if (optionParser.isSet(Opt.HELP)) {
						System.out.println(subCommand.getDetailedDescription());
						System.out.println("[available options]");
						System.out.println(subCommand.getOptionList());
						return;
					}
					else {
						// copy the rest of the command line arguments
						String[] subcommandArgs = new String[args.length - 1];
						for (int i = 1; i < args.length; i++)
							subcommandArgs[i - 1] = args[i];
						subCommand.execute(subcommandArgs);
						return;
					}
				}
				else
					System.err.println("unknown subcommand: " + subCommandName);
			}
			else {
				if (optionParser.isSet(Opt.HELP)) {
					// display help message
					System.out.println(getProgramInfo());
					BufferedReader helpReader = FileResource.open(UTGBShell.class, "help-message.txt");
					String line;
					while ((line = helpReader.readLine()) != null)
						System.out.println(line);

					// list command line options
					System.out.println(optionParser.helpMessage());

					// list all sub commands
					System.out.println("[sub commands]");
					for (String subCommandName : subCommandTable.keySet()) {
						UTGBShellSubCommand sc = subCommandTable.get(subCommandName);
						System.out.format("  %-10s\t%s", subCommandName, sc.getOneLinerDescription());
						System.out.println();
					}
					return;
				}

				if (optionParser.isSet(Opt.VERSION)) {
					System.out.println(getProgramInfo());
					return;
				}

			}

			// display a short help message
			System.out.println(getProgramInfo());
			System.out.println("type --help for a list of the available sub commands.");

		}
		catch (OptionParserException e) {
			System.err.println(e.getMessage());

		}
		catch (IOException e) {
			System.err.println(e.getMessage());
		}
		catch (UTGBShellException e) {
			System.err.println(e.getMessage());
		}

	}

	public static String getProgramInfo() {
		return "UTGB Shell: version " + getVersion() + " (" + getRevision() + ")";
	}

	public static String getRevision() {
		final String revision = "$Revision: 1435 $";
		return revision.substring(1, revision.length() - 1).trim();
	}

	public static String getVersion() {
		POM pom = new POM();
		// read the pom.xml file and extract the version information
		try {
			// load the pom.xml file copied as a resource
			InputStream pomIn = UTGBShell.class.getResourceAsStream("/pom.xml");
			if (pomIn != null) {
				BufferedReader pomReader = new BufferedReader(new InputStreamReader(pomIn));
				if (pomReader != null)
					BeanUtil.populateBeanWithXML(pom, pomReader);
			}
		}
		catch (XerialException e) {
			_logger.debug(e);
		}
		return pom.getVersion();
	}
}
